/*
    lm78.c - Part of a Linux module for reading sensor data.
    Copyright (c) 1997, 1998  Alexander Larsson <alla@lysator.liu.se>,
    Frodo Looijaard <frodol@dds.nl> and Philip Edelbrock <phil@netroedge.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/config.h>

/* The next define disables the redefinition of kernel_version. As it is
   already defined and initialized in lm78.c, this would result in twice
   defined symbols. We must include <linux/version.h> by hand now. */
#define __NO_VERSION__

/* Hack to allow the use of CONFIG_MODVERSIONS. In the kernel itself, this
   is arranged from the main Makefile. */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/module.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#include "lm78.h"
#include "main.h"

/* These are functions used only internally. */
static u8 lm78_read_value(u8 address);
static void lm78_write_value(u8 address, int value);

#define LMTYPE_UNKNOWN 0
#define LMTYPE_78 1
#define LMTYPE_78_J 2
#define LMTYPE_79 3
#define LMTYPE_WINB 4
static int lm_type = LMTYPE_UNKNOWN;

/* If set, we use the SMBus to access the LM78 */
static int lm78_over_smbus = 0;

/* This function initializes the LM78. If it is found, and initializing
   succeeds, 0 is returned; if not, an error value is returned. */
int LM78_Init(void)
{
  u8 res;
#ifdef DEBUG
  int temp;
#endif

  if (check_region(LM78_BASE, LM78_EXTENT))
    return -ENODEV;

  outb_p( LM78_DUMMY, LM78_ADDRESS_REG );
  res = inb_p(LM78_ADDRESS_REG);
  inb_p(LM78_DATA_REG); /* reset busy flag */

  if ((res & 0x7F) != LM78_DUMMY)
    return -EIO;

  request_region(LM78_BASE, LM78_EXTENT, "lm78");

  /* From now on, we will use lm78_write_value and lm78_read_value,
     as all other things are the same for SMBus and ISA bus access */

  /* Reset all except Watchdog values and last conversion values
     This sets fan-divs to 2, among others */
  lm78_write_value(LM78_CONFIG, 0x80 );

  /* Determine the LM version */

  /* Check for W83781D. Note the careful probing: we do not write anything
     to a register until we are already almost sure we have found the
     Winbond. */
  if (res = lm78_read_value(LM78_WINB_ID), (res == 0x5c) || (res == 0xa3) ) {
    res = lm78_read_value(LM78_WINB_BANK);
    lm78_write_value(LM78_WINB_BANK,res & 0x7f);
    if (lm78_read_value(LM78_WINB_ID) == 0xa3) {
      lm78_write_value(LM78_WINB_BANK,res | 0x80);
      if (lm78_read_value(LM78_WINB_ID) == 0x5c)
        lm_type = LMTYPE_WINB;
    }
    lm78_write_value(LM78_WINB_BANK,res);
  }

  if (lm_type == LMTYPE_UNKNOWN) {
    res = lm78_read_value(LM78_RESET) >> 6;
    if (res == 2)
       lm_type = LMTYPE_79;
    else if (res == 1)
       lm_type = LMTYPE_78_J;
    else if (res == 0)
       lm_type = LMTYPE_78;
   }

  /* Winbond initializations. Note that, according to docs, the Winbond-
     specific registers are only set to default values when the Master
     Reset pin is accessed - which is impossible to do through software.
     Here, I try to touch only the necessary stuff, and assume the BIOS
     took care of the more arcane settings. This may have to change,
     depending on input from users */
   if (lm_type == LMTYPE_WINB) {
     /* Put FAN3 at divide-by-2 */
     lm78_write_value(LM78_WINB_PIN,
                      (lm78_read_value(LM78_WINB_PIN) & 0x3f) | 0x40);
     /* Disable the power-on BEEP */
     lm78_write_value(LM78_WINB_FANCONT,
                      lm78_read_value(LM78_WINB_FANCONT) | 0x80);
     /* Make sure we use BANK0 from now on. BANK1/2 are for LM75-compatible
        sensors access through ISA, which we do not want right now */
     lm78_write_value(LM78_WINB_BANK,
                      (lm78_read_value(LM78_WINB_BANK) & 0xf8) | 0x01);
     /* Disable the beeping once and for all */
     lm78_write_value(LM78_WINB_BEEP2,lm78_read_value(LM78_WINB_BEEP2) & 0x7f);
   }


  /* Default limits --
        Note: a good default for IN0 and IN1 are values
        close to VID (not done here, yet) */

  LM78_Set(LM_PROBE_TEMP_MIN,0,0);
  LM78_Set(LM_PROBE_TEMP_MAX,0,50);

#ifdef LM78_IN0_STRING
  LM78_Set(LM_PROBE_IN_MIN,0,
         LM78_IN0_VAL - LM78_IN0_VAL * LM78_IN_PERCENTAGE / 100);
  LM78_Set(LM_PROBE_IN_MAX,0,
         LM78_IN0_VAL + LM78_IN0_VAL * LM78_IN_PERCENTAGE / 100);
#endif

#ifdef LM78_IN1_STRING
  LM78_Set(LM_PROBE_IN_MIN,1,
         LM78_IN1_VAL - LM78_IN1_VAL * LM78_IN_PERCENTAGE / 100);
  LM78_Set(LM_PROBE_IN_MAX,1,
         LM78_IN1_VAL + LM78_IN1_VAL * LM78_IN_PERCENTAGE / 100);
#endif

#ifdef LM78_IN2_STRING
  LM78_Set(LM_PROBE_IN_MIN,2,
         LM78_IN2_VAL - LM78_IN2_VAL * LM78_IN_PERCENTAGE / 100);
  LM78_Set(LM_PROBE_IN_MAX,2,
         LM78_IN2_VAL + LM78_IN2_VAL * LM78_IN_PERCENTAGE / 100);
#endif

#ifdef LM78_IN3_STRING
  LM78_Set(LM_PROBE_IN_MIN,3,
         LM78_IN3_VAL - LM78_IN3_VAL * LM78_IN_PERCENTAGE / 100);
  LM78_Set(LM_PROBE_IN_MAX,3,
         LM78_IN3_VAL + LM78_IN3_VAL * LM78_IN_PERCENTAGE / 100);
#endif

#ifdef LM78_IN4_STRING
  LM78_Set(LM_PROBE_IN_MIN,4,
         LM78_IN4_VAL - LM78_IN4_VAL * LM78_IN_PERCENTAGE / 100);
  LM78_Set(LM_PROBE_IN_MAX,4,
         LM78_IN4_VAL + LM78_IN4_VAL * LM78_IN_PERCENTAGE / 100);
#endif

#ifdef LM78_IN5_STRING
  LM78_Set(LM_PROBE_IN_MIN,5,
         LM78_IN5_VAL - LM78_IN5_VAL * LM78_IN_PERCENTAGE / 100);
  LM78_Set(LM_PROBE_IN_MAX,5,
         LM78_IN5_VAL + LM78_IN5_VAL * LM78_IN_PERCENTAGE / 100);
#endif

#ifdef LM78_IN6_STRING
  LM78_Set(LM_PROBE_IN_MIN,6,
         LM78_IN6_VAL - LM78_IN6_VAL * LM78_IN_PERCENTAGE / 100);
  LM78_Set(LM_PROBE_IN_MAX,6,
         LM78_IN6_VAL + LM78_IN6_VAL * LM78_IN_PERCENTAGE / 100);
#endif

#ifdef LM78_FAN1_STRING
  LM78_Set(LM_PROBE_FAN_MIN,1,LM78_FAN1_VAL);
#endif

#ifdef LM78_FAN2_STRING
  LM78_Set(LM_PROBE_FAN_MIN,2,LM78_FAN2_VAL);
#endif

#ifdef LM78_FAN3_STRING
  LM78_Set(LM_PROBE_FAN_MIN,3,LM78_FAN3_VAL);
#endif

  /* Start monitoring */
  LM78_WRITE(LM78_CONFIG, (LM78_READ(LM78_CONFIG) & 0xf7) | 0x01);

  if (lm_type == LMTYPE_UNKNOWN)
    printk("Unknown LM78 clone detected (treated as a plain LM78)\n");
  else {
    if (lm_type == LMTYPE_78) 
      printk("LM78");
    else if (lm_type == LMTYPE_78_J)
      printk("LM78-J");
    else if (lm_type == LMTYPE_79)
      printk("LM79");
    else if (lm_type == LMTYPE_WINB)
      printk("W83781D");
    printk(" detected and installed.\n");
  }
#ifdef DEBUG
  temp=LM78_SMBUS_READ(LM78_SMBUSADDR);
  if ((temp != 0) && (temp != -1))
        printk("LM78 Probe: Serial bus address 0x%X.\n",temp);
  else
        printk("LM78 Probe: failed!\n");
#endif
  return 0;
}

/* This function cleans up after the module. Only call it after a succesful
   LM78_Init call! */
void LM78_Cleanup(void)
{
   release_region(LM78_BASE, LM78_EXTENT);
}

/* This function tries to update the current sensor data and to put it
   in human-readable form in buf. It returns the number of characters
   written to buf. 
   TODO: Check length of buf. */
int LM78_Print_Values (char *buf)
{
  int len = 0;
  int v, ints1, ints2;

  /* Redundant test for LM_Sensor_Data.valid - better safe then sorry! */
  if (LM_Sensor_Data.valid) {
#define PRINT_SIGN(val) (((val) < 0)?'-':'+')
#define PRINT_VALUES(label, value, max, min, int) \
    sprintf(buf + len, label ":\t%c%d.%02dV\t(min = %c%d.%02dV,\t" \
            "max = %c%d.%02dV)\t%s\n", PRINT_SIGN(value), abs(value) / 100, abs(value) % 100, PRINT_SIGN(min), abs(min) / 100,\
            abs(min) % 100, PRINT_SIGN(max),abs(max) / 100, abs(max) % 100, int?"ALARM":"")

    ints1 = LM_Sensor_Data.interrupt_status[0];
    ints2 = LM_Sensor_Data.interrupt_status[1];

#ifdef LM78_IN0_STRING
    len += PRINT_VALUES(LM78_IN0_STRING, LM78_Get(LM_PROBE_IN,0), 
                        LM78_Get(LM_PROBE_IN_MAX,0), 
                        LM78_Get(LM_PROBE_IN_MIN,0), 
                        ints1 & 0x01);
#endif

#ifdef LM78_IN1_STRING
    len += PRINT_VALUES(LM78_IN1_STRING, LM78_Get(LM_PROBE_IN,1), 
                        LM78_Get(LM_PROBE_IN_MAX,1), 
                        LM78_Get(LM_PROBE_IN_MIN,1), 
                        ints1 & 0x02);
#endif

#ifdef LM78_IN2_STRING
    len += PRINT_VALUES(LM78_IN2_STRING, LM78_Get(LM_PROBE_IN,2), 
                        LM78_Get(LM_PROBE_IN_MAX,2), 
                        LM78_Get(LM_PROBE_IN_MIN,2), 
                        ints1 & 0x04);
#endif

#ifdef LM78_IN3_STRING
    len += PRINT_VALUES(LM78_IN3_STRING, LM78_Get(LM_PROBE_IN,3), 
                        LM78_Get(LM_PROBE_IN_MAX,3), 
                        LM78_Get(LM_PROBE_IN_MIN,3), 
                        ints1 & 0x08);
#endif

#ifdef LM78_IN4_STRING
    len += PRINT_VALUES(LM78_IN4_STRING, LM78_Get(LM_PROBE_IN,4), 
                        LM78_Get(LM_PROBE_IN_MAX,4), 
                        LM78_Get(LM_PROBE_IN_MIN,4), 
                        ints2 & 0x01);
#endif

#ifdef LM78_IN5_STRING
    len += PRINT_VALUES(LM78_IN5_STRING, LM78_Get(LM_PROBE_IN,5), 
                        LM78_Get(LM_PROBE_IN_MAX,5), 
                        LM78_Get(LM_PROBE_IN_MIN,5), 
                        ints2 & 0x02);
#endif

#ifdef LM78_IN6_STRING
    len += PRINT_VALUES(LM78_IN6_STRING, LM78_Get(LM_PROBE_IN,6), 
                        LM78_Get(LM_PROBE_IN_MAX,6), 
                        LM78_Get(LM_PROBE_IN_MIN,6), 
                        ints2 & 0x04);
#endif


#ifdef LM78_FAN1_STRING
    len += sprintf(buf+len,"%s:\t%d rpm   \t(min=%d rpm)\t%s\t%s\n", 
                   LM78_FAN1_STRING,
                   LM78_Get(LM_PROBE_FAN,1),LM78_Get(LM_PROBE_FAN_MIN,1),
                   ints1 & 0x40?"ALARM":"",
                   LM78_Get(LM_PROBE_FAN,1)==0?"(not connected?)":"");
#endif

#ifdef LM78_FAN2_STRING
    len += sprintf(buf+len, "%s:\t%d rpm   \t(min=%d rpm)\t%s\t%s\n", 
                   LM78_FAN2_STRING,
                   LM78_Get(LM_PROBE_FAN,2),LM78_Get(LM_PROBE_FAN_MIN,2),
                   ints1 & 0x80?"ALARM":"",
                   LM78_Get(LM_PROBE_FAN,2)==0?"(not connected?)":"");
#endif

#ifdef LM78_FAN3_STRING
    len += sprintf(buf+len, "%s:\t%d rpm   \t(min=%d rpm)\t%s\t%s\n", 
                   LM78_FAN3_STRING,
                   LM78_Get(LM_PROBE_FAN,3),LM78_Get(LM_PROBE_FAN_MIN,3),
                   ints2 & 0x08?"ALARM":"",
                   LM78_Get(LM_PROBE_FAN,3)==0?"(not connected?)":"");
#endif

    len += sprintf(buf+len, "Mainboard temperature: %d degrees (min = %d, "
                   "max = %d)\t%s\n", 
                   LM78_Get(LM_PROBE_TEMP,0), LM78_Get(LM_PROBE_TEMP_MIN,0),
                   LM78_Get(LM_PROBE_TEMP_MAX,0),ints1 & 0x10?"ALARM":"");

    len += sprintf(buf+len, "Other alarms:");

    if(ints1 & 0x20) len += sprintf(buf+len, " LM75");
    if(ints2 & 0x10) len += sprintf(buf+len, " Chassis_Intrusion");
    if (lm_type != LMTYPE_WINB) {
      if(ints2 & 0x20) len += sprintf(buf+len, " FIFO_Overflow");
      if(ints2 & 0x40) len += sprintf(buf+len, " SMI_IN");
    }
    if (! (lm_type==LMTYPE_WINB ? (ints1 & 0x20) | (ints2 & 0x10) :
                                  (ints1 & 0x20) | (ints2 & 0x70)))
      len += sprintf(buf+len," (none)");
    len += sprintf(buf+len, "\n");
        
#ifdef LM78_VID
    v = LM_Sensor_Data.vid;

    if (((lm_type == LMTYPE_78) || (lm_type == LMTYPE_78_J)) && (v == 0)) 
      len += sprintf(buf+len,"VID: probably not connected (says 3.5V)\n");
    else {
      if ((lm_type == LMTYPE_78) || (lm_type == LMTYPE_78_J))
        v |= 0x10;
   
      if (v <= 0x01)
          len += sprintf(buf+len, "VID:\t 2.0%dV\n", (0x01 - v) * 5);
      else if (v <= 0x04)
          len += sprintf(buf+len, "VID:\t 1.%dV\n", (0x04 - v) * 5 + 85);
      else if (v <= 0x0f)
          len += sprintf(buf+len, "VID:\t reserved (unknown) value\n");
      else if (v <= 0x15)
          len += sprintf(buf+len, "VID:\t 3.%dV\n", (0x15 - v));
      else if (v <= 0x1e)
          len += sprintf(buf+len, "VID:\t 2.%dV\n", (0x1e - v) + 1);
      else 
          len += sprintf(buf+len, "VID:\t No core installed.\n"); 
    } 
#endif
  }
  return len;
}


/* Read a value from the LM78. Theoretically, several drivers can use
   the LM78 at once by using this function. In practice, this is unlikely,
   but still it won't do any harm. */
u8 lm78_read_value(u8 address) 
{
  if (lm78_over_smbus)
    return LM78_SMBUS_READ(address);
  else {
    while (LM78_BUSY) {
      current->state = TASK_INTERRUPTIBLE;
      current->timeout = jiffies + 10;
      schedule();
    }
    return LM78_READ(address);
  }
}

void lm78_write_value(u8 address, int value)
{
#ifdef DEBUG
  printk("LM78: Write %d at address %d\n",value,address);
#endif
  if (lm78_over_smbus)
    LM78_SMBUS_WRITE(address,value);
  else {
    while (LM78_BUSY) {
      current->state = TASK_INTERRUPTIBLE;
      current->timeout = jiffies + 10;
      schedule();
    }
    LM78_WRITE(address,value);
  }
}
   

/* Reread the sensor information into struct LM_Sensor_Data. 
   We first check whether this is needed at all. */
void LM78_Update_Values(void)
{
  /* TODO: grab Spin lock on LM_Sensor_Data */

#ifdef LM78_IN0_STRING
  LM_Sensor_Data.in[0] = lm78_read_value(LM78_IN0);
  LM_Sensor_Data.in_max[0] = lm78_read_value(LM78_IN0_MAX);
  LM_Sensor_Data.in_min[0] = lm78_read_value(LM78_IN0_MIN);
#endif

#ifdef LM78_IN1_STRING
  LM_Sensor_Data.in[1] = lm78_read_value(LM78_IN1);
  LM_Sensor_Data.in_max[1] = lm78_read_value(LM78_IN1_MAX);
  LM_Sensor_Data.in_min[1] = lm78_read_value(LM78_IN1_MIN);
#endif

#ifdef LM78_IN2_STRING
  LM_Sensor_Data.in[2] = lm78_read_value(LM78_IN2);
  LM_Sensor_Data.in_max[2] = lm78_read_value(LM78_IN2_MAX);
  LM_Sensor_Data.in_min[2] = lm78_read_value(LM78_IN2_MIN);
#endif

#ifdef LM78_IN3_STRING
  LM_Sensor_Data.in[3] = lm78_read_value(LM78_IN3);
  LM_Sensor_Data.in_max[3] = lm78_read_value(LM78_IN3_MAX);
  LM_Sensor_Data.in_min[3] = lm78_read_value(LM78_IN3_MIN);
#endif

#ifdef LM78_IN4_STRING
  LM_Sensor_Data.in[4] = lm78_read_value(LM78_IN4);
  LM_Sensor_Data.in_max[4] = lm78_read_value(LM78_IN4_MAX);
  LM_Sensor_Data.in_min[4] = lm78_read_value(LM78_IN4_MIN);
#endif

#ifdef LM78_IN5_STRING
  LM_Sensor_Data.in[5] = lm78_read_value(LM78_IN5);
  LM_Sensor_Data.in_max[5] = lm78_read_value(LM78_IN5_MAX);
  LM_Sensor_Data.in_min[5] = lm78_read_value(LM78_IN5_MIN);
#endif

#ifdef LM78_IN6_STRING
  LM_Sensor_Data.in[6] = lm78_read_value(LM78_IN6);
  LM_Sensor_Data.in_max[6] = lm78_read_value(LM78_IN6_MAX);
  LM_Sensor_Data.in_min[6] = lm78_read_value(LM78_IN6_MIN);
#endif

#ifdef LM78_VID
  if (! LM_Sensor_Data.valid) {
    LM_Sensor_Data.vid = lm78_read_value(LM78_VID) & 0x0f;
    if ((lm_type == LMTYPE_79) || (lm_type == LMTYPE_WINB))
      LM_Sensor_Data.vid |= ((lm78_read_value(LM78_RESET) & 0x01) << 4);
  }
#endif

  LM_Sensor_Data.temp_mb = lm78_read_value(LM78_TEMPERATURE);
  LM_Sensor_Data.temp_mb_max = lm78_read_value(LM78_TEMPERATURE_MAX);
  LM_Sensor_Data.temp_mb_min = lm78_read_value(LM78_TEMPERATURE_MIN);

#ifdef LM78_FAN1_STRING
  LM_Sensor_Data.fan[0] = lm78_read_value(LM78_FAN1);
  LM_Sensor_Data.fan_min[0] = lm78_read_value(LM78_FAN1_MIN);
#endif
#ifdef LM78_FAN2_STRING
  LM_Sensor_Data.fan[1] = lm78_read_value(LM78_FAN2);
  LM_Sensor_Data.fan_min[1] = lm78_read_value(LM78_FAN2_MIN);
#endif
#ifdef LM78_FAN3_STRING
  LM_Sensor_Data.fan[2] = lm78_read_value(LM78_FAN3);
  LM_Sensor_Data.fan_min[2] = lm78_read_value(LM78_FAN3_MIN);
#endif


  LM_Sensor_Data.interrupt_status[0] = lm78_read_value(LM78_INTSTAT1);
  LM_Sensor_Data.interrupt_status[1] = lm78_read_value(LM78_INTSTAT2);

#ifdef DEBUG
  LM_Sensor_Data.lm78_smbus_addr = LM78_SMBUS_READ(LM78_SMBUSADDR);
#endif


  /* TODO: release Spin lock on LM_Sensor_Data */
}

/* This function will set a limit to a new value. It will scale it as
   needed. See lm78.h for the possible values of what and the scaling
   which is expected. */
void LM78_Set(int what, int nr, long value)
{
  u8 port;
  long write;

#ifdef DEBUG
  printk("LM78: LM78_Set %d %d %d\n",what,nr,value);
#endif

  switch(what) {
  case LM_PROBE_IN_MIN:
  case LM_PROBE_IN_MAX:
    switch(nr) {
    case 0:
      write=(value * 100000 / NORMAL_IN0) / 16;
      port=(what==LM_PROBE_IN_MIN?LM78_IN0_MIN:LM78_IN0_MAX);
      break;
    case 1:
      write=(value * 100000 / NORMAL_IN1) / 16;
      port=(what==LM_PROBE_IN_MIN?LM78_IN1_MIN:LM78_IN1_MAX);
      break;
    case 2:
      write=(value * 100000 / NORMAL_IN2) / 16;
      port=(what==LM_PROBE_IN_MIN?LM78_IN2_MIN:LM78_IN2_MAX);
      break;
    case 3:
      write=(value * 100000 / NORMAL_IN3) / 16;
      port=(what==LM_PROBE_IN_MIN?LM78_IN3_MIN:LM78_IN3_MAX);
      break;
    case 4:
      write=(value * 100000 / NORMAL_IN4) / 16;
      port=(what==LM_PROBE_IN_MIN?LM78_IN4_MIN:LM78_IN4_MAX);
      break;
    case 5:
      write=(value * 100000 / NORMAL_IN5) / 16;
      port=(what==LM_PROBE_IN_MIN?LM78_IN5_MIN:LM78_IN5_MAX);
      break;
    case 6:
      write=(value * 100000 / NORMAL_IN6) / 16;
      port=(what==LM_PROBE_IN_MIN?LM78_IN6_MIN:LM78_IN6_MAX);
      break;
    default:
      return;
    }
    if ((write < 0) || (write > 255)) 
      return;
    lm78_write_value(port,write);
    if (what == LM_PROBE_IN_MIN)
      LM_Sensor_Data.in_min[nr] = write;
    else
      LM_Sensor_Data.in_max[nr] = write;
    break;
  case LM_PROBE_FAN_MIN:
    switch(nr) {
    case 1:
      port=LM78_FAN1_MIN;
      break;
    case 2:
      port=LM78_FAN2_MIN;
      break;
    case 3:
      port=LM78_FAN3_MIN;
      break;
    default:
      return;
    }
    write=LM78_FAN_COUNT(value,2);
    if (write < 0 || write > 255) 
      return; /* value out of range */
    lm78_write_value(port,write);
    LM_Sensor_Data.fan_min[nr-1]=write;
    break;
  case LM_PROBE_TEMP_MIN:
  case LM_PROBE_TEMP_MAX:
    port=what==LM_PROBE_TEMP_MIN?LM78_TEMPERATURE_MIN:LM78_TEMPERATURE_MAX;
    if ((value >= 0) && (value <= 125))
      write=value;
    else if (value >= -55)
      write=value & 0xff;
    else /* value out of range */
      return;
    lm78_write_value(port,write);
    if (what == LM_PROBE_TEMP_MIN) 
      LM_Sensor_Data.temp_mb_min=write;
    else
      LM_Sensor_Data.temp_mb_max=write;
    break;
  default: /* what out of range */
      return;
  }
}      
       
/* We return 0 if what or nr are out-of-bounds. Note that you must call
   LM_Update_Data before you access this function! */
long LM78_Get(int what, int nr)
{
  int mult;

  switch(what) {
  case LM_PROBE_IN:
  case LM_PROBE_IN_MIN:
  case LM_PROBE_IN_MAX:
    switch(nr) {
      case 0:
        mult=NORMAL_IN0;
        break;
      case 1:
        mult=NORMAL_IN1;
        break;
      case 2:
        mult=NORMAL_IN2;
        break;
      case 3:
        mult=NORMAL_IN3;
        break;
      case 4:
        mult=NORMAL_IN4;
        break;
      case 5:
        mult=NORMAL_IN5;
        break;
      case 6:
        mult=NORMAL_IN6;
        break;
      default /* nr invalid */:
        return 0;
    }
    if (what == LM_PROBE_IN)
      return LM_Sensor_Data.in[nr] * 16 * mult / 100000;
    else if (what == LM_PROBE_IN_MIN)
      return LM_Sensor_Data.in_min[nr] * 16 * mult / 100000;
    else
      return LM_Sensor_Data.in_max[nr] * 16 * mult / 100000;
  case LM_PROBE_FAN:
  case LM_PROBE_FAN_MIN:
    if ((nr >= 1) && (nr <= 3)) {
      if (what == LM_PROBE_FAN)
        return LM78_RPM(LM_Sensor_Data.fan[nr-1],2);
      else
        return LM78_RPM(LM_Sensor_Data.fan_min[nr-1],2);
    } else
      return 0;
  case LM_PROBE_TEMP:
  case LM_PROBE_TEMP_MIN:
  case LM_PROBE_TEMP_MAX:
    if (what == LM_PROBE_TEMP)
      mult = LM_Sensor_Data.temp_mb;
    else if (what == LM_PROBE_TEMP_MIN)
      mult = LM_Sensor_Data.temp_mb_min;
    else 
      mult = LM_Sensor_Data.temp_mb_max;
    if (mult > 0x80)
      return mult-256;
    else
      return mult;
  default: /* what invalid */
    return 0;
  }
}
    
    
