/* -*- linux-c -*- */


/*
    lm80.c - Part of a Linux module for reading sensor data.
    Copyright (c) 1998  Philip Edelbrock <phil@netroedge.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/*
 *  LM80 sensor driver for LINUX
 *
 * Copyright (c) 1998 Philip Edelbrock
 * 
 * part of a module which provides a 
 * /proc/sensors file which shows
 * some information about your system.
 * Voltages, temperature and fan speeds
 * are currently reported.
 *
 *
 * please report any bugs/patches to me:
 *  phil@netroedge.com
 *
 *
 * Reference:
 *   National Semiconductor
 *   LM80 System Hardware Monitor with
 *   Two-Wire Interface, PDF, April 1997.
 *   http://www.national.com/ds/LM/LM80.pdf
 *
 *   Intel
 *   82371AB PCI-TO-ISA / IDE XCELERATOR (PIIX4), PDF, April 1997.
 *   ftp://ftp.intel.com
 *   
 *   System Management Bus Specification, Rev. 1.0, 1996.
 *   Contributors:
 *   Benchmarq Microelectronics Inc ., Duracell Inc.,
 *   Energizer Power Systems, Intel Corporation, Linear Technology Corporation,
 *   Maxim Integrated Products, Mitsubishi Electric Corporation,
 *   National Semiconductor Corporation, Toshiba Battery Co.,
 *   Varta Batterie AG, All rights reserved.
 *   http://www.sbs-forum.org/
 */

#include <linux/config.h>

/* The next define disables the redefinition of kernel_version. As it is
   already defined and initialized in lm78.c, this would result in twice
   defined symbols. We must include <linux/version.h> by hand now. */
#define __NO_VERSION__

/* Hack to allow the use of CONFIG_MODVERSIONS. In the kernel itself, this
   is arranged from the main Makefile. */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include "main.h"
#include "lm80.h"
#include "smbus.h"

/* Set to zero to attempt auto-detection, otherwise to the valid address) */
int lm80_smbus_addr=0;
int lm80_enabled=0;

/* Functions used only internally */


/* LM80 Initialization */
int LM80_Init () {

  int i;

/* seems like this should be in main.c
  if (! SMBus_Initialized)
    return -ENODEV;
 */

if (lm80_smbus_addr == 0) {
/* Let's find the LM80 (if we can) */
   for (i=0x2F; i >= 0x20; i--) { /* We decrement because the suggested address is 0x20 */
     if (SMBus_Read(i,1,BYTE_DATA) != -1) 
     	lm80_smbus_addr=i;
   }
} else {
/* Make sure given address is valid */
  if (SMBus_Read(i,1,BYTE_DATA) == -1)
     	lm80_smbus_addr=0;
  }
  if (lm80_smbus_addr == 0) { /* If nothing found, then report it */
  	printk("No LM80 detected.  LM80 code disabled.\n");
  	return -ENODEV;
   } else {
	lm80_enabled=1;
  	printk("LM80 detected at 0x%X.  LM80 code enabled.\n",lm80_smbus_addr);
	return 0;
   }
}

void LM80_Cleanup () {
/* seems like this should be in main.c
   SMBus_Cleanup();
*/
}

void LM80_Set(int reg, int value) {

   SMBus_Write(lm80_smbus_addr,reg | (value << 8),BYTE_DATA);
}

long LM80_Get(int what) {

 if (lm80_smbus_addr != 0) {
  if ( (what < 0x40) && (what > 0)) { /* make sure what is valid */
   return SMBus_Read(lm80_smbus_addr,what,BYTE_DATA) >> 8;
  } else {
   return 0;
  }
 }
 return 0;
}


int LM80_Print_Values (char *buf) {
 int i,j,value;
 int len=0;

/* For testing: I cheated and fetched the values as they were printed. */
 
 if (LM_Sensor_Data.valid) {

	
        value=LM80_Get(0);
        len += sprintf(buf+len, 
                      "LM80-%d config register: 0x%X\n",
                      lm80_smbus_addr - 0x20, value);
        value=LM80_Get(1);
        len += sprintf(buf+len, 
                      "LM80-%d interrupt register 1: 0x%X\n",
                      lm80_smbus_addr - 0x20, value);
        value=LM80_Get(2);
        len += sprintf(buf+len, 
                      "LM80-%d interrupt register 2: 0x%X\n",
                      lm80_smbus_addr - 0x20, value);
        value=LM80_Get(3);
        len += sprintf(buf+len, 
                      "LM80-%d interrupt mask register 1: 0x%X\n",
                      lm80_smbus_addr - 0x20, value);
        value=LM80_Get(4);
        len += sprintf(buf+len, 
                      "LM80-%d interrupt mask register 2: 0x%X\n",
                      lm80_smbus_addr - 0x20, value);
        value=LM80_Get(5);
        len += sprintf(buf+len, 
                      "LM80-%d fan_div/rst#/os# register: 0x%X\n",
                      lm80_smbus_addr - 0x20, value);
        value=LM80_Get(6);
        len += sprintf(buf+len, 
                      "LM80-%d os# config & temp resolution register: 0x%X\n",
                      lm80_smbus_addr - 0x20, value);

	for (i=0x20; i < 0x40; i+=0x10){
	  len += sprintf(buf+len, "\nLM80-%d value ram 0x%X-0x%X: ",lm80_smbus_addr - 0x20,
	  			i,i+0x10);
		for (j=0; j < 0x10;j++)
		  len += sprintf(buf+len," 0x%X", LM80_Get(i+j));
	}
	len += sprintf(buf+len, "\n");
  }
  return len;
}

void LM80_Update_Values (void) {

}


