/*
    sysctl.c - Part of a Linux module for reading sensor data.
    Copyright (c) 1997, 1998  Alexander Larsson <alla@lysator.liu.se>,
    Frodo Looijaard <frodol@dds.nl> and Philip Edelbrock <phil@netroedge.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/config.h>

/* The next define disables the redefinition of kernel_version. As it is
   already defined and initialized in lm78.c, this would result in twice
   defined symbols. We must include <linux/version.h> by hand now. */
#define __NO_VERSION__

/* Hack to allow the use of CONFIG_MODVERSIONS. In the kernel itself, this
   is arranged from the main Makefile. */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/version.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/proc_fs.h>
#include <linux/ctype.h>
#include <linux/types.h>   /* Needed for sysctl.h */
#include <linux/linkage.h> /* Needed for sysctl.h */
#include <linux/sysctl.h>
#include "lmsysctl.h"
#include "main.h"
#include "lm78.h"
#include "lm75.h"

#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,1,0))
#include <asm/uaccess.h>
#endif

static int lm_proc (ctl_table *ctl, int write, struct file * filp,
                    void *buffer, size_t *lenp);
static int lm_sysctl (ctl_table *table, int *name, int nlen,
                      void *oldval, size_t *oldlenp,
                      void *newval, size_t newlen,
                      void **context);
static void write_in(int nr, int nrels, long *results);
static void read_in(int nr, long *results);
static void write_fan(int nr, int nrels, long *results);
static void read_fan(int nr, long *results);
static void write_LM75(int nr, int nrels, long *results);
static void read_LM75(int nr, long *results);
static void write_temp(int nrels, long *results);
static void read_temp(long *results);
static void parse_elements(int *nrels, void *buffer, int bufsize, 
                           long *results, int magnitude);
static void write_elements(int nrels,void *buffer,int *bufsize,
                           long *results,int mag);

static ctl_table lm_lm_table[] = { 
#ifdef LM78_IN0_STRING
  { LMT_IN0, LM78_IN0_STRING, NULL, 0, 0644, NULL, &lm_proc, &lm_sysctl },
#endif
#ifdef LM78_IN1_STRING
  { LMT_IN1, LM78_IN1_STRING, NULL, 0, 0644, NULL, &lm_proc, &lm_sysctl },
#endif
#ifdef LM78_IN2_STRING
  { LMT_IN2, LM78_IN2_STRING, NULL, 0, 0644, NULL, &lm_proc, &lm_sysctl },
#endif
#ifdef LM78_IN3_STRING
  { LMT_IN3, LM78_IN3_STRING, NULL, 0, 0644, NULL, &lm_proc, &lm_sysctl },
#endif
#ifdef LM78_IN4_STRING
  { LMT_IN4, LM78_IN4_STRING, NULL, 0, 0644, NULL, &lm_proc, &lm_sysctl },
#endif
#ifdef LM78_IN5_STRING
  { LMT_IN5, LM78_IN5_STRING, NULL, 0, 0644, NULL, &lm_proc, &lm_sysctl },
#endif
#ifdef LM78_IN6_STRING
  { LMT_IN6, LM78_IN6_STRING, NULL, 0, 0644, NULL, &lm_proc, &lm_sysctl },
#endif
#ifdef LM78_FAN1_STRING
  { LMT_FAN1, LM78_FAN1_STRING, NULL, 0, 0644, NULL, &lm_proc, &lm_sysctl },
#endif
#ifdef LM78_FAN2_STRING
  { LMT_FAN2, LM78_FAN2_STRING, NULL, 0, 0644, NULL, &lm_proc, &lm_sysctl },
#endif
#ifdef LM78_FAN3_STRING
  { LMT_FAN3, LM78_FAN3_STRING, NULL, 0, 0644, NULL, &lm_proc, &lm_sysctl },
#endif
#ifdef LM75_0_STRING
  { LMT_LM75_0, LM75_0_STRING, NULL, 0, 0644, NULL, &lm_proc, &lm_sysctl },
#endif
#ifdef LM75_1_STRING
  { LMT_LM75_1, LM75_1_STRING, NULL, 0, 0644, NULL, &lm_proc, &lm_sysctl },
#endif
#ifdef LM75_2_STRING
  { LMT_LM75_2, LM75_2_STRING, NULL, 0, 0644, NULL, &lm_proc, &lm_sysctl },
#endif
#ifdef LM75_3_STRING
  { LMT_LM75_3, LM75_3_STRING, NULL, 0, 0644, NULL, &lm_proc, &lm_sysctl },
#endif
#ifdef LM75_4_STRING
  { LMT_LM75_4, LM75_4_STRING, NULL, 0, 0644, NULL, &lm_proc, &lm_sysctl },
#endif
#ifdef LM75_5_STRING
  { LMT_LM75_5, LM75_5_STRING, NULL, 0, 0644, NULL, &lm_proc, &lm_sysctl },
#endif
#ifdef LM75_6_STRING
  { LMT_LM75_6, LM75_6_STRING, NULL, 0, 0644, NULL, &lm_proc, &lm_sysctl },
#endif
#ifdef LM75_7_STRING
  { LMT_LM75_7, LM75_7_STRING, NULL, 0, 0644, NULL, &lm_proc, &lm_sysctl },
#endif
  { LMT_TEMP, "temp", NULL, 0, 0644, NULL, &lm_proc, &lm_sysctl },
  { 0 }
};

static ctl_table lm_dev_table[] = {
  { DEV_LM, "lm_sensors", NULL, 0, 0555, lm_lm_table },
  { 0 }
};

static ctl_table lm_root_table[] = {
  { CTL_DEV, "dev", NULL, 0, 0555, lm_dev_table },
  { 0 }
};

static struct ctl_table_header *lm_table_header;



int LM_Sysctl_Init(void) 
{
  lm_table_header = register_sysctl_table(lm_root_table,0);
  return lm_table_header?0:-ENOMEM;
}

void LM_Sysctl_Cleanup(void)
{
  if (lm_table_header)
    unregister_sysctl_table(lm_table_header);
}

int lm_proc (ctl_table *ctl, int write, struct file * filp,
             void *buffer, size_t *lenp)
{
  int nrels,mag;
  long results[3];

  /* If buffer is size 0, or we try to read when not at the start, we 
     return nothing. Note that I think writing when not at the start
     does not work either, but anyway, this is straight from the kernel
     sources. */
  if (!*lenp || (filp->f_pos && !write)) {
    *lenp = 0;
    return 0;
  }

  /* How many numbers are found within these files, and how to scale them? */
  switch (ctl->ctl_name) {
    case LMT_IN0: case LMT_IN1: case LMT_IN2: case LMT_IN3: case LMT_IN4:
    case LMT_IN5: case LMT_IN6:
      nrels=3;
      mag=2;
      break;
    case LMT_TEMP:
      nrels=3;
      mag=0;
      break;
    case LMT_FAN1: case LMT_FAN2: case LMT_FAN3:
      nrels=2;
      mag=0;
      break;
    case LMT_LM75_0: case LMT_LM75_1: case LMT_LM75_2: case LMT_LM75_3:
    case LMT_LM75_4: case LMT_LM75_5: case LMT_LM75_6: case LMT_LM75_7:
      nrels=3;
      mag=1;
      break;
    default: /* Should never be called */
      return -EINVAL;
  }
 
  /* OK, try writing stuff. */
  if (write) {
    parse_elements(&nrels,buffer,*lenp,results,mag);
    if (nrels == 0)
      return 0;
    switch (ctl->ctl_name) {
      case LMT_IN0: write_in(0,nrels,results); break;
      case LMT_IN1: write_in(1,nrels,results); break;
      case LMT_IN2: write_in(2,nrels,results); break;
      case LMT_IN3: write_in(3,nrels,results); break;
      case LMT_IN4: write_in(4,nrels,results); break;
      case LMT_IN5: write_in(5,nrels,results); break;
      case LMT_IN6: write_in(6,nrels,results); break;
      case LMT_FAN1: write_fan(1,nrels,results); break;
      case LMT_FAN2: write_fan(2,nrels,results); break;
      case LMT_FAN3: write_fan(3,nrels,results); break;
      case LMT_LM75_0: write_LM75(0,nrels,results);break;
      case LMT_LM75_1: write_LM75(1,nrels,results);break;
      case LMT_LM75_2: write_LM75(2,nrels,results);break;
      case LMT_LM75_3: write_LM75(3,nrels,results);break;
      case LMT_LM75_4: write_LM75(4,nrels,results);break;
      case LMT_LM75_5: write_LM75(5,nrels,results);break;
      case LMT_LM75_6: write_LM75(6,nrels,results);break;
      case LMT_LM75_7: write_LM75(7,nrels,results);break;
      case LMT_TEMP: write_temp(nrels,results); break;
      default: /* Should never be called */ return 0; break;
    }
    filp->f_pos += *lenp;
    return 0;
  } else { /* read */
    /* Update all values in LM_Sensor_Data */
    LM_Update_Data();

    /* Read the values to print into results */
    switch (ctl->ctl_name) {
      case LMT_IN0: read_in(0,results);break;
      case LMT_IN1: read_in(1,results);break;
      case LMT_IN2: read_in(2,results);break;
      case LMT_IN3: read_in(3,results);break;
      case LMT_IN4: read_in(4,results);break;
      case LMT_IN5: read_in(5,results);break;
      case LMT_IN6: read_in(6,results);break;
      case LMT_FAN1: read_fan(1,results);break;
      case LMT_FAN2: read_fan(2,results);break;
      case LMT_FAN3: read_fan(3,results);break;
      case LMT_LM75_0: read_LM75(0,results);break;
      case LMT_LM75_1: read_LM75(1,results);break;
      case LMT_LM75_2: read_LM75(2,results);break;
      case LMT_LM75_3: read_LM75(3,results);break;
      case LMT_LM75_4: read_LM75(4,results);break;
      case LMT_LM75_5: read_LM75(5,results);break;
      case LMT_LM75_6: read_LM75(6,results);break;
      case LMT_LM75_7: read_LM75(7,results);break;
      case LMT_TEMP: read_temp(results);break;
      default: /* Should never be called */ *lenp=0; return 0;
    }
    /* OK, print it now */
    write_elements(nrels,buffer,lenp,results,mag);
    filp->f_pos += *lenp;
    return 0;
  }
}

int lm_sysctl (ctl_table *table, int *name, int nlen, void *oldval, 
               size_t *oldlenp, void *newval, size_t newlen,
               void **context)
{
   /* No time to implement this yet. */
   return -ENOTDIR;
}
 
void write_in(int nr, int nrels, long *results) 
{
  if (nrels >= 1)
    LM78_Set(LM_PROBE_IN_MIN,nr,results[0]);
  if (nrels >= 2)
    LM78_Set(LM_PROBE_IN_MAX,nr,results[1]);
}

void read_in(int nr, long *results)
{
  results[0]=LM78_Get(LM_PROBE_IN_MIN,nr);
  results[1]=LM78_Get(LM_PROBE_IN_MAX,nr);
  results[2]=LM78_Get(LM_PROBE_IN,nr);
}

void write_fan(int nr, int nrels, long *results)
{
  if (nrels >= 1)
    LM78_Set(LM_PROBE_FAN_MIN,nr,results[0]);
}

void read_fan(int nr, long *results)
{
  results[0]=LM78_Get(LM_PROBE_FAN_MIN,nr);
  results[1]=LM78_Get(LM_PROBE_FAN,nr);
}

void write_LM75(int nr, int nrels, long *results)
{
  if (nrels >= 1)
    LM75_Set(LM_PROBE_LM75_MIN,nr,results[0]);
  if (nrels >= 2)
    LM75_Set(LM_PROBE_LM75_MAX,nr,results[1]);
}

void read_LM75(int nr, long *results)
{
  results[0]=LM75_Get(LM_PROBE_LM75_MIN,nr);
  results[1]=LM75_Get(LM_PROBE_LM75_MAX,nr);
  results[2]=LM75_Get(LM_PROBE_LM75_VALUE,nr);
}

void write_temp(int nrels, long *results)
{
  if (nrels >= 1)
    LM78_Set(LM_PROBE_TEMP_MIN,0,results[0]);
  if (nrels >= 2)
    LM78_Set(LM_PROBE_TEMP_MAX,0,results[1]);
}

void read_temp(long *results)
{
  results[0]=LM78_Get(LM_PROBE_TEMP_MIN,0);
  results[1]=LM78_Get(LM_PROBE_TEMP_MAX,0);
  results[2]=LM78_Get(LM_PROBE_TEMP,0);
}

/* nrels contains initially the maximum number of elements which can be
   put in results, and finally the number of elements actually put there.
   A magnitude of 1 will multiply everything with 10; etc.
   buffer, bufsize is the character buffer we read from and its length.
   results will finally contain the parsed integers. 

   Buffer should contain several reals, separated by whitespace. A real
   has the following syntax:
     [ Minus ] Digit* [ Dot Digit* ] 
   (everything between [] is optional; * means zero or more).
   When the next character is unparsable, everything is skipped until the
   next whitespace.

   WARNING! This is tricky code. I have tested it, but there may still be
            hidden bugs in it, even leading to crashes and things!
*/

void parse_elements(int *nrels, void *buffer, int bufsize, 
                    long *results, int magnitude)
{
  int maxels,min,mag;
  long res;
  char nextchar=0;

  maxels = *nrels;
  *nrels = 0;

  while (bufsize && (*nrels < maxels)) {

    /* Skip spaces at the start */
    while (bufsize && isspace((int) get_user((char *) buffer))) {
      bufsize --;
      ((char *) buffer)++;
    }

    /* Well, we may be done now */
    if (! bufsize)
      return;

    /* New defaults for our result */
    min = 0;
    res = 0;
    mag = magnitude;

    /* Check for a minus */
    if (get_user((char *) buffer) == '-') {
      min=1;
      bufsize--;
      ((char *) buffer)++;
    }

    /* Digits before a decimal dot */
    while (bufsize && (isdigit( (int) (nextchar=get_user((char *) buffer))))) {
      res = res * 10 + nextchar - '0';
      bufsize--;
      ((char *) buffer)++;
    }

    /* If mag < 0, we must actually divide here! */
    while (mag < 0) {
      res = res / 10;
      mag++;
    }

    if (bufsize && (nextchar == '.')) {
      /* Skip the dot */
      bufsize--;
      ((char *) buffer)++;
  
      /* Read digits while they are significant */
      while(bufsize && (mag > 0) && 
            (isdigit((int) (nextchar=get_user((char *) buffer))))) {
        res = res * 10 + nextchar - '0';
        mag--;
        bufsize--;
        ((char *) buffer)++;
      }
    }
    /* If we are out of data, but mag > 0, we need to scale here */
    while (mag > 0) {
      res = res * 10;
      mag --;
    }

    /* Skip everything until we hit whitespace */
    while(bufsize && (!isspace((int) (nextchar=get_user((char *) buffer))))) {
      bufsize --;
      ((char *) buffer) ++;
    }

    /* Put res in results */
    results[*nrels] = (min?-1:1)*res;
    (*nrels)++;
  }    
  
  /* Well, there may be more in the buffer, but we need no more data. 
     Ignore anything that is left. */
  return;
}
    
void write_elements(int nrels,void *buffer,int *bufsize,long *results,
                    int magnitude)
{
  #define BUFLEN 20
  char BUF[BUFLEN+1]; /* An individual representation should fit in here! */
  char printfstr[10];
  int nr=0;
  int buflen,mag,times;
  int curbufsize=0;

  while ((nr < nrels) && (curbufsize < *bufsize)) {
    mag=magnitude;

    if (nr != 0) {
      put_user(' ', (char *) buffer);
      curbufsize ++;
      ((char *) buffer) ++;
    }

    /* Fill BUF with the representation of the next string */
    if (mag <= 0) {

      buflen=sprintf(BUF,"%ld",results[nr]);
      if (buflen < 0) { /* Oops, a sprintf error! */
        *bufsize=0;
        return;
      }
      while ((mag < 0) && (buflen < BUFLEN)) {
        BUF[buflen++]='0';
        mag++;
      }
      BUF[buflen]=0;
    } else {
      times=1;
      for (times=1; mag-- > 0; times *= 10);
      if (results[nr] < 0) {
        BUF[0] = '-';
        buflen = 1;
      } else
        buflen=0;
      strcpy(printfstr,"%ld.%0Xld");
      printfstr[6]=magnitude+'0';
      buflen+=sprintf(BUF+buflen,printfstr,abs(results[nr])/times,
                      abs(results[nr])%times);
      if (buflen < 0) { /* Oops, a sprintf error! */
        *bufsize=0;
        return;
      }
    }

    /* Now copy it to the user-space buffer */
    if (buflen + curbufsize > *bufsize)
      buflen=*bufsize-curbufsize;
    copy_to_user(buffer,BUF,buflen);
    curbufsize += buflen;
    (char *) buffer += buflen;

    nr ++;
  }
  if (curbufsize < *bufsize) {
    put_user('\n', (char *) buffer);
    curbufsize ++;
  }
  *bufsize=curbufsize;
}
